/*
#ident	"@(#)conf/lib:RELEASE-3_2_0_108:sockets.h,v 1.3 1997/05/09 20:13:45 woods Exp"
 */

/*
 * sockets.h:
 *   header files that must be included to use BSD-style sockets.
 *
 * These includes can be overridden by the SOCKET_INCLUDES variable
 * in the EDITME or conf/os files.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#ifdef HAVE_BIND
# include <arpa/nameser.h>
# include <resolv.h>
#endif
