#include <sys/types.h>
#include <stdio.h>
#include "defs.h"
#include "smail.h"
#include "config.h"
#ifndef DEPEND
# include "extern.h"
#endif

#ifdef HAVE_LIBWRAP

/*
 * These badly named global variables are used by TCP Wrappers libwrap.a and
 * cannot yet be changed in the smail config file.
 */
/* 
 * XXX Some systems, such as NetBSD-1.3, require these definitions if -lwrap is
 * included in the link, even if nothing else from that library would otherwise
 * be included.
 */
int allow_severity = LIBWRAP_ALLOW_FACILITY | LIBWRAP_ALLOW_SEVERITY;
int deny_severity = LIBWRAP_DENY_FACILITY | LIBWRAP_DENY_SEVERITY;

#endif
