/*
#ident	"@(#)smail/compat:RELEASE-3_2_0_108:dummy.c,v 1.3 1997/08/08 03:44:06 woods Exp"
 */

/*
 * Provide a dummy reference so there is at least one item in
 * compat.a.  This will keep ld from complaining about empty libraries
 * and such.
 */
int
dummy_reference()
{
    return 1;
}
