/* @(#) log.h,v 1.5 1998/04/11 23:30:29 woods Exp */

/*
 *    Copyright (C) 1987, 1988 Ronald S. Karr and Landon Curt Noll
 *    Copyright (C) 1992  Ronald S. Karr
 * 
 * See the file COPYING, distributed with smail, for restriction
 * and warranty information.
 */

/*
 * log.h:
 *	interface file for routines in log.c
 */

/* bit-flag macros used in log.c */
#define WRITE_LOG_NONE	0x0000		/* write_log sends to no log */
#define WRITE_LOG_SYS	0x0001		/* write_log sends to system log */
#define WRITE_LOG_CONS	0x0002		/* write_log sends to console */
#define WRITE_LOG_MLOG	0x0004		/* write_log sends to per-msg log */
#define WRITE_LOG_PANIC	0x0008		/* write_log sends to panic log */
#define WRITE_LOG_TTY	0x0010		/* write_log sends to stderr */

#define WRITE_LOG_LAST	WRITE_LOG_TTY	/* the one with the highest bit set */
