/*
#ident	"@(#)smail/src/routers:RELEASE-3_2_0_108:rtlib.h,v 1.6 1998/11/20 01:02:29 woods Exp"
 */

/*
 * rtlib.h - interface file for routines in rtlib.c
 */

/* flag values passed from rt[dv]_standard to the driver routines */
#define RT_VERIFY	0x0001		/* Verify only */

/* external functions in the rtlib.c file */
void rtd_standard();
